"""
Our extensions to the loopy type system
"""
from dune.perftool.generation import (function_mangler,
                                      include_file,
                                      )

from loopy.symbolic import FunctionIdentifier
from loopy.types import NumpyType
from loopy import CallMangleInfo


import numpy as np


class VCLTypeRegistry:
    pass


def _populate_vcl_type_registry():
    VCLTypeRegistry.types = {}
    VCLTypeRegistry.names = {}

    # The base types that we are working with!
    for base_name, base_type, abbrev in [('float', np.float32, 'f'),
                                         ('double', np.float64, 'd'),
                                         ]:
        # The vector width in bits we are considering!
        for vector_bits in [128, 256, 512]:
            # Calculate the vector lane width
            count = vector_bits // (np.dtype(base_type).itemsize * 8)

            # Define the name of this vector type
            name = "Vec{}{}".format(count, abbrev)

            # Construct the numpy dtype!
            fieldnames = tuple("x" + str(i) for i in range(count))
            dtype = np.dtype(dict(names=fieldnames,
                                  formats=[base_type] * count,
                                  )
                             )

            VCLTypeRegistry.types[np.dtype(base_type), count] = dtype
            VCLTypeRegistry.names[dtype] = name


_populate_vcl_type_registry()


def get_vcl_type_size(nptype, register_size=256):
    return register_size // (np.dtype(nptype).itemsize * 8)


def get_vcl_type(nptype, register_size=None, vec_size=None):
    assert register_size or vec_size
    assert not(register_size and vec_size)

    if not vec_size:
        vec_size = register_size // (np.dtype(nptype).itemsize * 8)

    return VCLTypeRegistry.types[np.dtype(nptype), vec_size]


@function_mangler
def vcl_mul_add(knl, func, arg_dtypes):
    if func == "mul_add":
        vcl = lp.types.NumpyType(get_vcl_type(np.float64, register_size=256))
        return lp.CallMangleInfo("mul_add", (vcl), (vcl, vcl, vcl))
